from PAMI.sequentialPatternMining import abstract as _ab


class prefixSpan(_ab._frequentPatterns):
    """
    Attributes:
    ----------
            iFile : str
                Input file name or path of the input file
            oFile : str
                Name of the output file or the path of output file
            minSup: float or int or str
                The user can specify minSup either in count or proportion of database size.
                If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
                Otherwise, it will be treated as float.
                Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
            sep : str
                This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
                However, the users can override their default separator.
            startTime:float
                To record the start time of the mining process
            endTime:float
                To record the completion time of the mining process
            finalPatterns: dict
                Storing the complete set of patterns in a dictionary variable
            memoryUSS : float
                To store the total amount of USS memory consumed by the program
            memoryRSS : float
                To store the total amount of RSS memory consumed by the program
            Database : list
                To store the transactions of a database in list
            patternBuffer : list
                To store the items and used to help while saving the patterns
            containsItemSetsWithMultipleItems : boolean
                Check whether the dataset contains multiple items or not

    Methods:
    -------
        startMine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        savePatterns(oFile)
            Complete set of frequent patterns will be loaded in to a output file
        getPatternsAsDataFrame()
            Complete set of frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        minSupAbs()
            Convert the percentage value minimumSupport to the float or integer value
        save()
            save the patterns in the finalPatterns dictionary
        save1()
            save the patterns in the finalPatterns dictionary
        scanDataBase()
            scans the database and assign the database to the variable in the form of a list
        PseudoSequence()
            returns the parameters in the form of the list
        findSequenceContainingItems()
            returns the dictionary in which contains the single items with their sequence Ids
        prefixSpanWithSingleItems()
            removes the infrequent items and start the process to find the patterns with single items
        prefixSpanWithMultipleItems()
            removes the infrequent items and start the process to find the patterns with multiple items
        buildProjectedDataBaseSingleItems()
            build the projected database for the single items
        buildProjectedDataBaseFirstTimeMultipleItems()
            build the projected database for the multiple items
        findAllFrequentPairsSingleItems()
            calculates the all possible pattern pairs for the single items
        findAllFrequentPairs()
            calculate the all possible pattern pairs for the multiple items
        recursionSingleItems()
            recursive process will be done for the pattern if it is frequent for the single items
        recursion()
            recursive process will be done for the pattern if it is frequent for the multiple items
    """
    _minSup = float()
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _patternBuffer = [None] * 50
    _containsItemSetsWithMultipleItems = False

    def _minSupAbs(self, file, support):
        """
        calculates and return the integer or float value for minimumSupport
        :param file: input file
        :param support: minimumSupport value as percentage
        """
        with open(file, 'r') as f:
            read = f.readlines()
        self.minSup = int(_ab._math.ceil(support * len(read)))

    def _save(self, item, support):
        """
        saves the frequent sequential pattern  in the 'output.txt' file
        :param item: item
        :param support: support value
        """
        a = ""
        a += str(item)
        a += " -1"
        self._finalPatterns[a] = str(support)

    def _save1(self, lastBufPos, pseudoSeq):
        """
        saves the frequent sequential pattern  in the 'output.txt' file
        :param lastBufPos: buffer position to save the sequential frequent pattern
        :param pseudoSeq: sequences containing the item
        """
        i = 0
        a = ""
        while i <= lastBufPos:
            a += str(self._patternBuffer[i])
            a += " "
            i += 1
        if self._patternBuffer[lastBufPos] != -1:
            a += str(-1)
        self._finalPatterns[a] = str(len(pseudoSeq))

    def _scanDataBase(self):
        """
        scans the database and stores the items in the list
        :return: returns the database in the form of the list
        """

        with open(self._iFile, 'r') as f:
            read = f.readlines()
            lis = []
            for i in read:
                li = list(map(int, i.split()))
                lis.append(li)
        print(lis, type(lis))
        return lis

    @staticmethod
    def _PseudoSequence(seqId, indFirItem):
        """
        returns the parameters in the form of the list
        :param seqId: sequence id in the database
        :param indFirItem: index of the item in the sequence
        :return: return the parameters in the form of the list
        """
        return [seqId, indFirItem]

    def _findSequenceContainItems(self, data):
        """
        calculates the items and their sequences contains single items in the database
        :param data: input data in the form of the list
        :return: returns the items contains single items in the database
        """
        sequenceID = _ab._dd(list)
        for i in range(len(data)):
            sequence = data[i]
            itemCount = 0
            for j in sequence:
                if j >= 0:
                    sequenceIDs = sequenceID[j]
                    if len(sequenceIDs) == 0 or sequenceIDs[len(sequenceIDs) - 1] != i:
                        sequenceIDs.append(i)
                    itemCount += 1
                    if itemCount > 1:
                        self.containsItemSetsWithMultipleItems = True
                elif j == -1:
                    itemCount = 0
        return sequenceID

    def _prefixSpanWithSingleItems(self, seqId, data):
        """
        remove the infrequent items and starts the process to find frequent the sequential patterns
        :param seqId: sequence containing the items and their sequences
        :param data: input data in the form of the list
        """
        for i in range(len(data)):
            seq = data[i]
            curPos = 0
            lis = []
            for j in seq:
                if j >= 0:
                    if j in seqId:
                        isFreq = len(seqId[j]) >= self.minSup
                        if isFreq:
                            lis.append(j)
                            curPos += 1
                elif j == -2:
                    if curPos > 0:
                        lis.append(-2)
                        data[i] = lis
                        continue
                    else:
                        data[i] = []
        self.Database = data
        for i in list(sorted(seqId)):
            if len(seqId[i]) >= self.minSup:
                self._save(i, len(seqId[i]))
                self._patternBuffer[0] = i
                projDataB = self._buildProjectedDatabaseSingleItems(i, seqId[i])
                self._recursionSingleItems(self._patternBuffer, projDataB, 0)

    def _prefixSpanWithMultipleItems(self, seqId, data):
        """
        remove the infrequent items and starts the process to find frequent the sequential patterns
        :param seqId: sequence containing the items and their sequences
        :param data: input data in the form of the list
        """
        for i in range(len(data)):
            seq = data[i]
            curPos = 0
            curItemCount = 0
            lis = []
            for j in seq:
                if j >= 0:
                    if j in seqId:
                        isFreq = len(seqId[j]) >= self._minSup
                        if isFreq:
                            lis.append(j)
                            curPos += 1
                            curItemCount += 1
                elif j == -1:
                    if curItemCount > 0:
                        lis.append(-1)
                        curPos += 1
                        curItemCount = 0
                elif j == -2:
                    if curPos > 0:
                        lis.append(-2)
                        data[i] = lis
                        continue
                    else:
                        data[i] = []
        self._Database = data
        for i in list(sorted(seqId)):
            if len(seqId[i]) >= self.minSup:
                self._save(i, len(seqId[i]))
                self._patternBuffer[0] = i
                projDataB = self._buildProjectedDatabaseFirstTimeMultipleItems(i, seqId[i])
                self._recursion(self._patternBuffer, projDataB, 0)

    def _buildProjectedDatabaseSingleItems(self, item, seqId):
        """
        build the projected database for the item
        :param item: item in the database
        :param seqId: sequence of the item in the database
        :return: return the projected database in the form of the list
        """
        proDataBase = []
        for i in seqId:
            k = self.Database[i]
            for j in range(len(k)):
                token = k[j]
                if token != -2:
                    if token == item:
                        if k[j + 1] != -2:
                            proData = self._PseudoSequence(i, j + 1)
                            proDataBase.append(proData)
                            continue
        return proDataBase

    def _buildProjectedDatabaseFirstTimeMultipleItems(self, item, seqId):
        """
        build the projected database for the item
        :param item: item in the database
        :param seqId: sequence of the item in the database
        :return: return the projected database in the form of the list
        """
        proDataBase = []
        for i in seqId:
            k = self.Database[i]
            for j in range(len(k)):
                token = k[j]
                if token != -2:
                    if token == item:
                        isEndOfSeq = (k[j + 1] == -1 and k[j + 2] == -2)
                        if not isEndOfSeq:
                            proData = self._PseudoSequence(i, j + 1)
                            proDataBase.append(proData)
                            continue
        return proDataBase

    def _findAllFrequentPairsSingleItems(self, dataBase):
        """
        calculates the all possible pattern pairs for the items
        :param dataBase: projected database
        :return: return the dictionary containing item and their values
        """
        mapItemPseudoSeq = _ab._dd(list)
        for i in range(len(dataBase)):
            seqId = dataBase[i][0]
            seq = self.Database[seqId]
            for j in range(dataBase[i][1], len(seq)):
                token = seq[j]
                if token >= 0:
                    listSeq = mapItemPseudoSeq[token]
                    ok = True
                    if len(listSeq) > 0:
                        ok = listSeq[len(listSeq) - 1][0] != seqId
                    if ok:
                        listSeq.append(self._PseudoSequence(seqId, j + 1))
        return mapItemPseudoSeq

    def _findAllFrequentPairs(self, dataBase, lastBufPos):
        """
        calculates the all possible pattern pairs for the items
        :param dataBase: projected database
        :param lastBufPos: pattern buffer position to save the pattern
        :return: return the dictionary containing item and their values
        """
        pseudoSeqMP = {}
        pseudoSeqMPIP = {}
        mapsPairs = _ab._dd(list)
        mapPairsInPostfix = _ab._dd(list)
        firstPosOfLasItemBuf = lastBufPos
        while lastBufPos > 0:
            firstPosOfLasItemBuf -= 1
            if firstPosOfLasItemBuf < 0 or self._patternBuffer[firstPosOfLasItemBuf] == -1:
                firstPosOfLasItemBuf += 1
                break
        posToMatch = firstPosOfLasItemBuf
        for i in range(len(dataBase)):
            seqId = dataBase[i][0]
            seq = self.Database[seqId]
            prevItem = seq[dataBase[i][1] - 1]
            curItemSetIsPostfix = (prevItem != -1)
            isFirstItemSet = True
            for j in range(dataBase[i][1], len(seq)):
                token = seq[j]
                if token >= 0:
                    pair = token
                    if curItemSetIsPostfix:
                        if pair not in list(pseudoSeqMPIP):
                            pseudoSeqMPIP[pair] = []
                    else:
                        if pair not in list(pseudoSeqMP):
                            pseudoSeqMP[pair] = []
                    if curItemSetIsPostfix:
                        oldPair = mapPairsInPostfix[pair]
                    else:
                        oldPair = mapsPairs[pair]
                    if not oldPair:
                        if curItemSetIsPostfix:
                            mapPairsInPostfix[pair] = pair
                        else:
                            mapsPairs[pair] = pair
                    else:
                        pair = oldPair
                    ok = True
                    if curItemSetIsPostfix:
                        if len(pseudoSeqMPIP[pair]) > 0:
                            ok = pseudoSeqMPIP[pair][len(pseudoSeqMPIP[pair]) - 1][0] != seqId
                        if ok:
                            pseudoSeqMPIP[pair].append(self._PseudoSequence(seqId, j + 1))
                    else:
                        if len(pseudoSeqMP[pair]) > 0:
                            ok = pseudoSeqMP[pair][len(pseudoSeqMP[pair]) - 1][0] != seqId
                        if ok:
                            pseudoSeqMP[pair].append(self._PseudoSequence(seqId, j + 1))
                    if curItemSetIsPostfix and not isFirstItemSet:
                        pair = token
                        oldPair = mapsPairs[pair]
                        if not oldPair:
                            mapsPairs[pair] = pair
                        else:
                            pair = oldPair
                        ok = True
                        if pair not in list(pseudoSeqMP):
                            pseudoSeqMP[pair] = []
                        if len(pseudoSeqMP[pair]) > 0:
                            ok = pseudoSeqMP[pair][len(pseudoSeqMP[pair]) - 1][0] != seqId
                        if ok:
                            pseudoSeqMP[pair].append(self._PseudoSequence(seqId, j + 1))
                    if curItemSetIsPostfix is False and self._patternBuffer[posToMatch] is token:
                        posToMatch += 1
                        if posToMatch > lastBufPos:
                            curItemSetIsPostfix = True
                elif token == -1:
                    isFirstItemSet = False
                    curItemSetIsPostfix = False
                    posToMatch = firstPosOfLasItemBuf
        return [mapsPairs, mapPairsInPostfix, pseudoSeqMP, pseudoSeqMPIP]

    def _recursionSingleItems(self, patternBuff, dataBase, lastBuffPo):
        """
        recursion of the items if the pattern is frequent and finds the other pattern from that sequence
        :param patternBuff: pattern buffer we have taken
        :param dataBase: projected database
        :param lastBuffPo: buffer position in the pattern buffer
        """
        itemPseudoSeq = self._findAllFrequentPairsSingleItems(dataBase)
        if itemPseudoSeq:
            for i in list(sorted(itemPseudoSeq)):
                if len(itemPseudoSeq[i]) >= self.minSup:
                    patternBuff[lastBuffPo + 1] = -1
                    patternBuff[lastBuffPo + 2] = i
                    self._save1(lastBuffPo + 2, itemPseudoSeq[i])
                    self._recursionSingleItems(patternBuff, itemPseudoSeq[i], lastBuffPo + 2)

    def _recursion(self, patternBuff, dataBase, lastBuffPo):
        """
        recursion of the items if the pattern is frequent and finds the other pattern from that sequence
        :param patternBuff: pattern buffer we have taken
        :param dataBase: projected database
        :param lastBuffPo: buffer position in the pattern buffer
        """
        mapsPairs = self._findAllFrequentPairs(dataBase, lastBuffPo)
        if mapsPairs:
            for i in list(sorted(mapsPairs[1])):
                if len(mapsPairs[3][i]) >= self._minSup:
                    newBufferPos = lastBuffPo
                    newBufferPos += 1
                    patternBuff[newBufferPos] = i
                    self._save1(newBufferPos, mapsPairs[3][i])
                    self._recursion(patternBuff, mapsPairs[3][i], newBufferPos)
            for i in list(sorted(mapsPairs[0])):
                if len(mapsPairs[2][i]) >= self._minSup:
                    newBufferPos = lastBuffPo
                    newBufferPos += 1
                    patternBuff[newBufferPos] = -1
                    newBufferPos += 1
                    patternBuff[newBufferPos] = i
                    self._save1(newBufferPos, mapsPairs[2][i])
                    self._recursion(patternBuff, mapsPairs[2][i], newBufferPos)

    def startMine(self):
        """
        mining of the patterns starts from here
        """
        self._startTime = _ab._time.time()
        dat = self._scanDataBase()
        seq = self._findSequenceContainItems(dat)
        self._minSupAbs(self._iFile, self._minSup)
        if self._containsItemSetsWithMultipleItems:
            self._prefixSpanWithMultipleItems(seq, dat)
        else:
            self._prefixSpanWithSingleItems(seq, dat)
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss

    def getMemoryUSS(self):
        """
        Total USS memory consumed by the program will be retrieved by this function
        :return:
        """
        return self._memoryUSS

    def getMemoryRSS(self):
        """
        Total RSS memory consumed by the program will be retrieved by this function
        :return: return the memory
        """
        return self._memoryRSS

    def getRuntime(self):
        """
        Total amount of runtime taken by the program will be retrieved from this function
        :return: returns the time
        """
        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """
        complete set of patterns will be stored in the dataframe by this function
        :return: returns the dataframe
        """
        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b])
            dataFrame = _ab._pd.DataFrame(data, columns=['SequentialPatterns', 'Support'])
        return dataFrame

    def getPatterns(self):
        """
        Complete set of frequent patterns will be retrieved by this function
        :return: returns the frequent patterns dictionary
        """
        return self._finalPatterns

    def savePatterns(self, oFile):
        """
        save the frequent patterns in the output file
        :param oFile: output file which we saved the patterns
        """
        writer = open(oFile, 'w+')
        for x, y in self._finalPatterns.items():
            st = x + " # SUP " + y
            writer.write("%s \n" % st)


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 4 or len(_ab._sys.argv) == 5:
        if len(_ab._sys.argv) == 5:
            _ap = prefixSpan(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        if len(_ab._sys.argv) == 4:
            _ap = prefixSpan(_ab._sys.argv[1], _ab._sys.argv[3])
        _ap.startMine()
        _Patterns = _ap.getPatterns()
        print("Total number of Frequent Patterns:", len(_Patterns))
        _ap.savePatterns(_ab._sys.argv[2])
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
    else:
        '''l = [600,700,800, 900, 1000]
        for i in l:
            ap = prefixSpan('/Users/Likhitha/Downloads/PrefixSpan/small.txt', 1, ' ')
            ap.startMine()
            Patterns = ap.getPatterns()
            print("Total number of Frequent Patterns:", len(Patterns))
            ap.savePatterns('/Users/Likhitha/Downloads/output')
            memUSS = ap.getMemoryUSS()
            print("Total Memory in USS:", memUSS)
            memRSS = ap.getMemoryRSS()
            print("Total Memory in RSS", memRSS)
            run = ap.getRuntime()
            print("Total ExecutionTime in ms:", run)'''
        print("Error! The number of input parameters do not match the total number of parameters provided")




import statistics
import pandas as pd
import validators
from urllib.request import urlopen

class temporalDatabaseStats:
    """
    temporalDatabaseStats is class to get stats of database.

        Attributes:
        ----------
        inputFile : file
            input file path
        database : dict
            store time stamp and its transaction
        lengthList : list
            store size of all transaction
        timeStampCount : dict
            number of transactions per time stamp
        periodList : list
            all period list in the database
        sep : str
            separator in file. Default is tab space.

        Methods:
        -------
        run()
            execute readDatabase function
        readDatabase()
            read database from input file
        getDatabaseSize()
            get the size of database
        getMinimumTransactionLength()
            get the minimum transaction length
        getAverageTransactionLength()
            get the average transaction length. It is sum of all transaction length divided by database length.
        getMaximumTransactionLength()
            get the maximum transaction length
        getStandardDeviationTransactionLength()
            get the standard deviation of transaction length
        getSortedListOfItemFrequencies()
            get sorted list of item frequencies
        getSortedListOfTransactionLength()
            get sorted list of transaction length
        save(data, outputFile)
            store data into outputFile
        getMinimumPeriod()
            get the minimum period
        getAveragePeriod()
            get the average period
        getMaximumPeriod()
            get the maximum period
        getStandardDeviationPeriod()
            get the standard deviation period
        getNumberOfTransactionsPerTimestamp()
            get number of transactions per time stamp. This time stamp range is 1 to max period.
    """

    def __init__(self, inputFile, sep='\t'):
        """
        :param inputFile: input file name or path
        :type inputFile: str
        :param sep:
        """
        self.inputFile = inputFile
        self.database = {}
        self.lengthList = []
        self.timeStampCount = {}
        self.periodList = []
        self.sep = sep

    def run(self):
        self.readDatabase()

    def readDatabase(self):
        """
        read database from input file and store into database and size of each transaction.
        And store the period between transactions as list
        """
        numberOfTransaction = 0
        if isinstance(self.inputFile, pd.DataFrame):
            if self.inputFile.empty:
                print("its empty..")
            i = self.inputFile.columns.values.tolist()
            if 'ts' in i and 'Transactions' in i:
                self.database = self.inputFile.set_index('ts').T.to_dict(orient='records')[0]
            if 'ts' in i and 'Patterns' in i:
                self.database = self.inputFile.set_index('ts').T.to_dict(orient='records')[0]
            self.timeStampCount = self.inputFile.groupby('ts').count().T.to_dict(orient='records')[0]

        if isinstance(self.inputFile, str):
            if validators.url(self.inputFile):
                data = urlopen(self.inputFile)
                for line in data:
                    numberOfTransaction += 1
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self.sep)]
                    temp = [x for x in temp if x]
                    self.database[numberOfTransaction] = temp[1:]
                    self.timeStampCount[int(temp[0])] = self.timeStampCount.get(int(line[0]), 0)
                    self.timeStampCount[int(temp[0])] += 1
            else:
                try:
                    with open(self.inputFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            numberOfTransaction += 1
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self.sep)]
                            temp = [x for x in temp if x]
                            if len(temp) > 0:
                                self.database[numberOfTransaction] = temp[1:]
                                self.timeStampCount[int(temp[0])] = self.timeStampCount.get(int(line[0]), 0)
                                self.timeStampCount[int(temp[0])] += 1
                except IOError:
                    print("File Not Found")
                    quit()
        self.lengthList = [len(s) for s in self.database.values()]
        timeStampList = sorted(list(self.database.keys()))
        preTimeStamp = 0
        for ts in timeStampList:
            self.periodList.append(int(ts)-preTimeStamp)
            preTimeStamp = ts

        # for line in self.Database:
        #     numberOfTransaction += 1
        #     self.database[numberOfTransaction] = line[1:]
        #     self.timeStampCount[int(line[0])] = self.timeStampCount.get(int(line[0]), 0)
        #     self.timeStampCount[int(line[0])] += 1
        # self.lengthList = [len(s) for s in self.database.values()]
        # timeStampList = sorted(list(self.timeStampCount.keys()))
        # preTimeStamp = 0
        # for ts in timeStampList:
        #     self.periodList.append(int(ts)-preTimeStamp)
        #     preTimeStamp = ts

    def getDatabaseSize(self):
        """
        get the size of database
        :return: data base size
        """
        return len(self.database)

    def getMinimumTransactionLength(self):
        """
        get the minimum transaction length
        :return: minimum transaction length
        """
        return min(self.lengthList)

    def getAverageTransactionLength(self):
        """
        get the average transaction length. It is sum of all transaction length divided by database length.
        :return: average transaction length
        """
        totalLength = sum(self.lengthList)
        return totalLength / len(self.database)

    def getMaximumTransactionLength(self):
        """
        get the maximum transaction length
        :return: maximum transaction length
        """
        return max(self.lengthList)

    def getStandardDeviationTransactionLength(self):
        """
        get the standard deviation transaction length
        :return: standard deviation transaction length
        """
        return statistics.pstdev(self.lengthList)

    def getVarianceTransactionLength(self):
        """
        get the variance transaction length
        :return: variance transaction length
        """
        return statistics.variance(self.lengthList)

    def getSparsity(self):
        # percentage of 0 dense dataframe
        """
        get the sparsity of database
        :return: database sparsity
        """
        matrixSize = self.getDatabaseSize()*len(self.getSortedListOfItemFrequencies())
        return (matrixSize - sum(self.getSortedListOfItemFrequencies().values())) / matrixSize

    def getTotalNumberOfItems(self):
        """
        get the number of items in database.
        :return: number of items
        """
        return len(self.getSortedListOfItemFrequencies())

    def getSortedListOfItemFrequencies(self):
        """
        get sorted list of item frequencies
        :return: item frequencies
        """
        itemFrequencies = {}
        for tid in self.database:
            for item in self.database[tid]:
                itemFrequencies[item] = itemFrequencies.get(item, 0)
                itemFrequencies[item] += 1
        return {k: v for k, v in sorted(itemFrequencies.items(), key=lambda x:x[1], reverse=True)}

    def getTransanctionalLengthDistribution(self):
        """
        get transaction length
        :return: transaction length
        """
        transactionLength = {}
        for length in self.lengthList:
            transactionLength[length] = transactionLength.get(length, 0)
            transactionLength[length] += 1
        return {k: v for k, v in sorted(transactionLength.items(), key=lambda x: x[0])}

    def save(self, data, outputFile):
        """
        store data into outputFile
        :param data: input data
        :type data: dict
        :param outputFile: output file name or path to store
        :type outputFile: str
        """
        with open(outputFile, 'w') as f:
            for key, value in data.items():
                f.write(f'{key}\t{value}\n')

    def getMinimumPeriod(self):
        """
        get the minimum period
        :return: minimum period
        """
        return  min(self.periodList)

    def getAveragePeriod(self):
        """
        get the average period. It is sum of all period divided by number of period.
        :return: average period
        """
        totalPeriod = sum(self.periodList)
        return totalPeriod / len(self.periodList)

    def getMaximumPeriod(self):
        """
        get the maximum period
        :return: maximum period
        """
        return max(self.periodList)

    def getStandardDeviationPeriod(self):
        """
        get the standard deviation period
        :return: standard deviation period
        """
        return statistics.pstdev(self.periodList)

    def getNumberOfTransactionsPerTimestamp(self):
        """
        get number of transactions per time stamp
        :return: number of transactions per time stamp as dict
        """
        maxTS = max(list(self.timeStampCount.keys()))
        return {ts: self.timeStampCount.get(ts,0) for ts in range(1, maxTS+1)}

if __name__ == '__main__':
    data = {'ts': [1, 1, 3, 4, 5, 6, 7],

            'Transactions': [['a', 'd', 'e'], ['b', 'a', 'f', 'g', 'h'], ['b', 'a', 'd', 'f'], ['b', 'a', 'c'],
                             ['a', 'd', 'g', 'k'],

                             ['b', 'd', 'g', 'c', 'i'], ['b', 'd', 'g', 'e', 'j']]}

    data = pd.DataFrame.from_dict(data)
    obj = temporalDatabaseStats('/Users/Likhitha/Downloads/datasets-2/temporalDatabases/JDSA_datasets/temporal_FAA_merged.txt', '\t')
    import PAMI.extras.graph.plotLineGraphFromDictionary as plt
    obj.run()
    print(f'Database size : {obj.getDatabaseSize()}')
    print(f'Minimum Transaction Size : {obj.getMinimumTransactionLength()}')
    print(f'Average Transaction Size : {obj.getAverageTransactionLength()}')
    print(f'Maximum Transaction Size : {obj.getMaximumTransactionLength()}')
    print(f'Standard Deviation Transaction Size : {obj.getStandardDeviationTransactionLength()}')
    print(f'Variance : {obj.getVarianceTransactionLength()}')
    print(f'Sparsity : {obj.getSparsity()}')
    print(f'Number of items : {obj.getTotalNumberOfItems()}')
    print(f'Minimum period : {obj.getMinimumPeriod()}')
    print(f'Average period : {obj.getAveragePeriod()}')
    print(f'Maximum period : {obj.getMaximumPeriod()}')
    itemFrequencies = obj.getSortedListOfItemFrequencies()
    transactionLength = obj.getTransanctionalLengthDistribution()
    numberOfTransactionPerTimeStamp = obj.getNumberOfTransactionsPerTimestamp()
    # obj.save(itemFrequencies, 'itemFrequency.csv')
    # obj.save(transactionLength, 'transactionSize.csv')
    # obj.save(numberOfTransactionPerTimeStamp, 'numberOfTransaction.csv')
    plt.plotLineGraphFromDictionary(itemFrequencies, 100, 'itemFrequencies', 'item rank', 'frequency')
    plt.plotLineGraphFromDictionary(transactionLength, 100, 'transaction length', 'transaction length', 'frequency')
    plt.plotLineGraphFromDictionary(numberOfTransactionPerTimeStamp, 100)



#  Copyright (C)  2021 Rage Uday Kiran
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
import sys

from PAMI.periodicCorrelatedPattern import abstract as _ab


_maxPer = float()
_minAllConf = float()
_minSup = float()
_maxPerAllConf = float()
_frequentList = {}
_lno = int()


class _Node(object):
    """
        A class used to represent the node of frequentPatternTree

        Attributes:
        ----------
            item : int or None
                Storing item of a node
            timeStamps : list
                To maintain the timestamps of a database at the end of the branch
            parent : node
                To maintain the parent of every node
            children : list
                To maintain the children of a node

        Methods:
        -------
            addChild(itemName)
                Storing the children to their respective parent nodes
        """

    def __init__(self, item, children):
        """ Initializing the Node class

        :param item: Storing the item of a node
        :type item: int or None
        :param children: To maintain the children of a node
        :type children: dict
        """

        self.item = item
        self.children = children
        self.parent = None
        self.timeStamps = []

    def addChild(self, node):
        """ To add the children to a node

            :param node: parent node in the tree
        """

        self.children[node.item] = node
        node.parent = self


class _Tree(object):
    """
        A class used to represent the frequentPatternGrowth tree structure

        Attributes:
        ----------
            root : Node
                Represents the root node of the tree
            summaries : dictionary
                Storing the nodes with same item name
            info : dictionary
                Stores the support of the items


        Methods:
        -------
            addTransactions(Database)
                Creating transaction as a branch in frequentPatternTree
            getConditionalPatterns(Node)
                Generates the conditional patterns from tree for specific node
            conditionalTransaction(prefixPaths,Support)
                Takes the prefixPath of a node and support at child of the path and extract the frequent patterns from
                prefixPaths and generates prefixPaths with items which are frequent
            remove(Node)
                Removes the node from tree once after generating all the patterns respective to the node
            generatePatterns(Node)
                Starts from the root node of the tree and mines the periodic-frequent patterns

        """

    def __init__(self):
        self.root = _Node(None, {})
        self.summaries = {}
        self.info = {}

    def addTransaction(self, transaction, tid):
        """     Adding a transaction into tree

                :param transaction: To represent the complete database
                :type transaction: list
                :param tid: To represent the timestamp of a database
                :type tid: list
                :return: pfp-growth tree
        """

        currentNode = self.root
        for i in range(len(transaction)):
            if transaction[i] not in currentNode.children:
                newNode = _Node(transaction[i], {})
                currentNode.addChild(newNode)
                if transaction[i] in self.summaries:
                    self.summaries[transaction[i]].append(newNode)
                else:
                    self.summaries[transaction[i]] = [newNode]
                currentNode = newNode
            else:
                currentNode = currentNode.children[transaction[i]]
        currentNode.timeStamps = currentNode.timeStamps + tid

    def getConditionalPatterns(self, alpha, pattern):
        """Generates all the conditional patterns of a respective node

            :param alpha: To represent a Node in the tree
            :type alpha: Node
            :return: A tuple consisting of finalPatterns, conditional pattern base and information
        """
        finalPatterns = []
        finalSets = []
        for i in self.summaries[alpha]:
            set1 = i.timeStamps
            set2 = []
            while i.parent.item is not None:
                set2.append(i.parent.item)
                i = i.parent
            if len(set2) > 0:
                set2.reverse()
                finalPatterns.append(set2)
                finalSets.append(set1)
        finalPatterns, finalSets, info = self.conditionalDatabases(finalPatterns, finalSets, pattern)
        return finalPatterns, finalSets, info

    @staticmethod
    def generateTimeStamps(node):
        """To get the timestamps of a node

        :param node: A node in the tree
        :return: Timestamps of a node
        """

        finalTimeStamps = node.timeStamps
        return finalTimeStamps

    def removeNode(self, nodeValue):
        """ Removing the node from tree

            :param nodeValue: To represent a node in the tree
            :type nodeValue: node
            :return: Tree with their nodes updated with timestamps
        """

        for i in self.summaries[nodeValue]:
            i.parent.timeStamps = i.parent.timeStamps + i.timeStamps
            del i.parent.children[nodeValue]

    def getTimeStamps(self, alpha):
        """ To get all the timestamps of the nodes which share same item name

            :param alpha: Node in a tree
            :return: Timestamps of a  node
        """
        temporary = []
        for i in self.summaries[alpha]:
            temporary += i.timeStamps
        return temporary

    @staticmethod
    def getSupportAndPeriod(timeStamps, pattern):
        """To calculate the periodicity and support

        :param timeStamps: Timestamps of an item set
        :return: support, periodicity
        """

        global _minSup, _minAllConf, _maxPer, _maxPerAllConf, _frequentList, _lno
        timeStamps.sort()
        cur = 0
        per = list()
        sup = 0
        for j in range(len(timeStamps)):
            per.append(timeStamps[j] - cur)
            cur = timeStamps[j]
            sup += 1
        per.append(_lno - cur)
        if len(per) == 0:
            return [0, 0, 0, 0]
        l = []
        for i in pattern:
            l.append(_frequentList[i][0])
        l1 = []
        for i in pattern:
            l1.append(_frequentList[i][1])
        conf = sup/max(l)
        perConf = max(per)/min(l1)
        #print(pattern, timeStamps, l, l1, sup, max(per), conf, perConf)
        return [sup, max(per), conf, perConf]

    def conditionalDatabases(self, conditionalPatterns, conditionalTimeStamps, pattern):
        """ It generates the conditional patterns with periodic-frequent items

            :param conditionalPatterns: conditionalPatterns generated from conditionPattern method of a respective node
            :type conditionalPatterns: list
            :param conditionalTimeStamps: Represents the timestamps of a conditional patterns of a node
            :type conditionalTimeStamps: list
            :returns: Returns conditional transactions by removing non-periodic and non-frequent items
        """

        global _maxPer, _minSup
        temp = pattern
        pat = []
        timeStamps = []
        data1 = {}
        for i in range(len(conditionalPatterns)):
            for j in conditionalPatterns[i]:
                if j in data1:
                    data1[j] = data1[j] + conditionalTimeStamps[i]
                else:
                    data1[j] = conditionalTimeStamps[i]
        updatedDictionary = {}
        for m in data1:
            updatedDictionary[m] = self.getSupportAndPeriod(data1[m], pattern + [m])
        updatedDictionary = {k: v for k, v in updatedDictionary.items() if v[0] >= _minSup and v[1] <= _maxPer}
        count = 0
        for p in conditionalPatterns:
            p1 = [v for v in p if v in updatedDictionary]
            trans = sorted(p1, key=lambda x: (updatedDictionary.get(x)[0], -x), reverse=True)
            if len(trans) > 0:
                pat.append(trans)
                timeStamps.append(conditionalTimeStamps[count])
            count += 1
        return pat, timeStamps, updatedDictionary

    def generatePatterns(self, prefix):
        """ Generates the patterns

            :param prefix: Forms the combination of items
            :type prefix: list
            :returns: yields patterns with their support and periodicity
        """
        global _minSup, _minAllConf, _maxPer, _maxPerAllConf
        for i in sorted(self.summaries, key=lambda x: (self.info.get(x)[0], -x)):
            pattern = prefix[:]
            pattern.append(i)
            print(pattern, self.info[i][0], self.info[i][1], self.info[i][2], self.info[i][3])
            if self.info[i][0] >= _minSup and self.info[i][1] <= _maxPer and self.info[i][2] >= _minAllConf and self.info[i][3] <= _maxPerAllConf:
                yield pattern, self.info[i]
                patterns, timeStamps, info = self.getConditionalPatterns(i, pattern)
                conditionalTree = _Tree()
                conditionalTree.info = info.copy()
                for pat in range(len(patterns)):
                    conditionalTree.addTransaction(patterns[pat], timeStamps[pat])
                if len(patterns) > 0:
                    for q in conditionalTree.generatePatterns(pattern):
                        yield q
            self.removeNode(i)


class EPCPGrowth(_ab._periodicCorrelatedPatterns):
    """ EPCPGrowth is an algorithm to discover periodic-Correlaterd patterns in a temporal database.

    Reference:
    --------
        http://www.tkl.iis.u-tokyo.ac.jp/new/uploads/publication_file/file/897/Venkatesh2018_Chapter_DiscoveringPeriodic-Correlated.pdf

    Attributes:
    ----------
        iFile : file
            Name of the Input file or path of the input file
        oFile : file
            Name of the output file or path of the output file
        minSup: int or float or str
            The user can specify minSup either in count or proportion of database size.
            If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
            Otherwise, it will be treated as float.
            Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
        minAllConf: int or float or str
            The user can specify minAllConf either in count or proportion of database size.
            If the program detects the data type of minAllConf is integer, then it treats minAllCOnf is expressed in count.
            Otherwise, it will be treated as float.
            Example: minAllCOnf=10 will be treated as integer, while minAllConf=10.0 will be treated as float
        maxPer: int or float or str
            The user can specify maxPer either in count or proportion of database size.
            If the program detects the data type of maxPer is integer, then it treats maxPer is expressed in count.
            Otherwise, it will be treated as float.
            Example: maxPer=10 will be treated as integer, while maxPer=10.0 will be treated as float
        maxPerAllConf: int or float or str
            The user can specify maxPerAllConf either in count or proportion of database size.
            If the program detects the data type of maaxPerAllConf is integer, then it treats maxPerAllConf is expressed in count.
            Otherwise, it will be treated as float.
            Example: maxPerAllConf=10 will be treated as integer, while maxPerAllConf=10.0 will be treated as float
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
            However, the users can override their default separator.
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        Database : list
            To store the transactions of a database in list
        mapSupport : Dictionary
            To maintain the information of item and their frequency
        lno : int
            To represent the total no of transaction
        tree : class
            To represents the Tree class
        itemSetCount : int
            To represents the total no of patterns
        finalPatterns : dict
            To store the complete patterns

    Methods:
    -------
        startMine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        savePatterns(oFile)
            Complete set of periodic-frequent patterns will be loaded in to a output file
        getPatternsAsDataFrame()
            Complete set of periodic-frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        creatingItemSets(fileName)
            Scans the dataset and stores in a list format
        PeriodicFrequentOneItem()
            Extracts the one-periodic-frequent patterns from database
        updateDatabases()
            Update the database by removing aperiodic items and sort the Database by item decreased support
        buildTree()
            After updating the Database, remaining items will be added into the tree by setting root node as null
        convert()
            to convert the user specified value

        Executing the code on terminal:
        -------
        Format:
        ------
        python3 PFPGrowth.py <inputFile> <outputFile> <minSup> <maxPer>

        Examples:
        --------
        python3 PFPGrowth.py sampleTDB.txt patterns.txt 0.3 0.4   (minSup and maxPer will be considered in percentage of database
        transactions)

        python3 PFPGrowth.py sampleTDB.txt patterns.txt 3 4     (minSup and maxPer will be considered in support count or frequency)

        Sample run of importing the code:
        -------------------

            from PAMI.periodicCorrelatedPattern.basic import EPCPGrowth as alg

            obj = alg.EPCPGrowth(iFile, minSup, minAllCOnf, maxPer, maxPerAllConf)

            obj.startMine()

            periodicCorrelatedPatterns = obj.getPatterns()

            print("Total number of Periodic Frequent Patterns:", len(periodicCorrelatedPatterns))

            obj.savePatterns(oFile)

            Df = obj.getPatternsAsDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

        Credits:
        -------
            The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.\n

    """
    _startTime = float()
    _endTime = float()
    _minSup = str()
    _minAllCOnf = float()
    _maxPer = float()
    _maxPerAllConf = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _rank = {}
    _rankedUp = {}
    _lno = 0

    def _creatingItemSets(self):
        """
            Storing the complete transactions of the database/input file in a database variable


        """
        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            data, ts = [], []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'TS' in i:
                ts = self._iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self._iFile['Transactions'].tolist()
            for i in range(len(data)):
                tr = [ts[i][0]]
                tr = tr + data[i]
                self._Database.append(tr)

        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()


    def _periodicFrequentOneItem(self):
        """ Calculates the support of each item in the database and assign ranks to the items
            by decreasing support and returns the frequent items list

            :returns: return the one-length periodic frequent patterns
        """
        global _frequentList
        data = {}
        for tr in self._Database:
            for i in range(1, len(tr)):
                if tr[i] not in data:
                    data[tr[i]] = [int(tr[0]), int(tr[0]), 1]
                else:
                    data[tr[i]][0] = max(data[tr[i]][0], (int(tr[0]) - data[tr[i]][1]))
                    data[tr[i]][1] = int(tr[0])
                    data[tr[i]][2] += 1
        for key in data:
            data[key][0] = max(data[key][0], abs(len(self._Database) - data[key][1]))
        data = {k: [v[2], v[0], 1, 1] for k, v in data.items() if v[0] <= self._maxPer and v[2] >= self._minSup}
        pfList = [k for k, v in sorted(data.items(), key=lambda x: (x[1][0], x[0]), reverse=True)]
        self._rank = dict([(index, item) for (item, index) in enumerate(pfList)])
        for x, y in self._rank.items():
            _frequentList[y] = data[x]
        return data, pfList

    def _updateDatabases(self, dict1):
        """ Remove the items which are not frequent from database and updates the database with rank of items

            :param dict1: frequent items with support
            :type dict1: dictionary
            :return: Sorted and updated transactions
            """
        list1 = []
        for tr in self._Database:
            list2 = [int(tr[0])]
            for i in range(1, len(tr)):
                if tr[i] in dict1:
                    list2.append(self._rank[tr[i]])
            if len(list2) >= 2:
                basket = list2[1:]
                basket.sort()
                list2[1:] = basket[0:]
                list1.append(list2)
        return list1

    @staticmethod
    def _buildTree(data, info):
        """ It takes the database and support of an each item and construct the main tree by setting root node as a null

            :param data: it represents the one Databases in database
            :type data: list
            :param info: it represents the support of each item
            :type info: dictionary
            :return: returns root node of tree
        """

        rootNode = _Tree()
        rootNode.info = info.copy()
        for i in range(len(data)):
            set1 = [data[i][0]]
            rootNode.addTransaction(data[i][1:], set1)
        return rootNode

    def _savePeriodic(self, itemSet):
        """ To convert the ranks of items in to their original item names

            :param itemSet: frequent pattern
            :return: frequent pattern with original item names
        """
        t1 = str()
        for i in itemSet:
            t1 = t1 + self._rankedUp[i] + " "
        return t1

    def _convert(self, value):
        """
        To convert the given user specified value

        :param value: user specified value
        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value

    def startMine(self):
        """ Mining process will start from this function
        """

        global _minSup, _maxPer, _minAllConf, _maxPerAllConf, _lno
        self._startTime = _ab._time.time()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self._minSup is None:
            raise Exception("Please enter the Minimum Support")
        self._creatingItemSets()
        self._minSup = self._convert(self._minSup)
        self._minAllConf = float(self._minAllConf)
        self._maxPer = self._convert(self._maxPer)
        self._maxPerAllConf = float(self._maxPerAllConf)
        _minSup, _minAllConf, _maxPer, _maxPerAllConf, _lno = self._minSup, self._minAllConf,  self._maxPer, self._maxPerAllConf, len(self._Database)
        print(_minSup, _minAllConf, _maxPer, _maxPerAllConf)
        if self._minSup > len(self._Database):
            raise Exception("Please enter the minSup in range between 0 to 1")
        generatedItems, pfList = self._periodicFrequentOneItem()
        updatedDatabases = self._updateDatabases(generatedItems)
        for x, y in self._rank.items():
            self._rankedUp[y] = x
        info = {self._rank[k]: v for k, v in generatedItems.items()}
        Tree = self._buildTree(updatedDatabases, info)
        patterns = Tree.generatePatterns([])
        self._finalPatterns = {}
        for i in patterns:
            sample = self._savePeriodic(i[0])
            self._finalPatterns[sample] = i[1]
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Periodic Frequent patterns were generated successfully using PFPGrowth algorithm ")

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final periodic-frequent patterns in a dataframe

        :return: returning periodic-frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b[0], b[1], b[2], b[3]])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support', 'allConf', 'Periodicity', 'maxPerAllConf'])
        return dataFrame

    def savePatterns(self, outFile):
        """Complete set of periodic-frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x + ":" + str(y[0]) + ":" + str(y[1]) + ":" + str(y[2]) + ":" + str(y[3])
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """ Function to send the set of periodic-frequent patterns after completion of the mining process

        :return: returning periodic-frequent patterns
        :rtype: dict
        """
        return self._finalPatterns


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 7 or len(_ab._sys.argv) == 8:
        if len(_ab._sys.argv) == 8:
            _ap = EPCPGrowth(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5], sys.argv[6], sys.argv[7])
        if len(_ab._sys.argv) == 7:
            _ap = EPCPGrowth(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], sys.argv[5], sys.argv[6])
        _ap.startMine()
        _Patterns = _ap.getPatterns()
        print("Total number of Patterns:", len(_Patterns))
        _ap.savePatterns(_ab._sys.argv[2])
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
    else:
        '''ap = EPCPGrowth('/Users/Likhitha/Downloads/SPP_sample.txt', 2, 0.6, 6, 1.5,  ' ')
        ap.startMine()
        Patterns = ap.getPatterns()
        print(ap.getPatternsAsDataFrame())
        print("Total number of Frequent Patterns:", len(Patterns))
        ap.savePatterns('/Users/Likhitha/Downloads/EPCP_output.txt')
        memUSS = ap.getMemoryUSS()
        print("Total Memory in USS:", memUSS)
        memRSS = ap.getMemoryRSS()
        print("Total Memory in RSS", memRSS)
        run = ap.getRuntime()
        print("Total ExecutionTime in ms:", run)'''
        print("Error! The number of input parameters do not match the total number of parameters provided")




from PAMI.frequentPattern.pyspark import abstract as _ab


class parallelApriori(_ab._frequentPatterns):
    """
    Attributes
    ----------
        iFile : file
            Input file name or path of the input file
        oFile : file
            Name of the output file or the path of output file
        minSup : float
            minSup is a proportion of database size.
        numWorkers : int
            the number of workers
        finalPatterns : dict
            Storing the complete set of patterns in a dictionary variable
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program

    Methods:
    -------
        startMine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        storePatternsInFile(oFile)
            Complete set of frequent patterns will be loaded in to a output file
        getPatternsInDataFrame()
            Complete set of frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        getAllFrequentPatterns(database,frequentItemsets)
            This function generates all frequent patterns
        genCandidateItemsets(frequentPatterns, length)
            This function generates candidate patterns from the frequentPatterns
        Mapper(transaction,candidateItemsets)
            This function map each itemset of candidateItemsets to (itemset,1) if itemset is in transaction

    Executing the code on terminal:
    -------------------------------

        Format:
        ------
            python3 parallelApriori.py <inputFile> <outputFile> <minSup> <numWorkers>

        Examples:
        ---------
                python3 Apriori.py sampleDB.txt patterns.txt 10.0 10  (minSup will be considered in times of minSup and count of database transactions)
                python3 Apriori.py sampleDB.txt patterns.txt 10 8    (minSup will be considered in support count or frequency)
    """

    _minSup = float()
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _numWorkers = int()

    def __init__(self, iFile, minSup, numWorkers, sep = '\t'):
         super().__init__(iFile, float(minSup), int(numWorkers), sep)

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function
        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function
        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process
        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe
        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def savePatterns(self, outFile):
        """
        Complete set of frequent patterns will be loaded in to a output file
        :param outFile: name of the output file
        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = str(x) + " : " + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """
        Function to send the set of frequent patterns after completion of the mining process
        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    @staticmethod
    def _Mapper(transaction, candidateItemsets):
        """
        Map each candidate itemset of candidateItemsets to (itemset,1) if a candidate itemset is in transaction

        :param transaction: a transaction of database
        :type transaction: set
        :param candidateItemsets: candidate item sets
        :type candidateItemsets: list
        :return:set
        """

        candidates = set()
        for itemset in candidateItemsets:
            if set(itemset).issubset(transaction):
                candidates.add((itemset, 1))
        return candidates

    @staticmethod
    def _genCandidateItemsets(frequentPatterns, length):
        """
        Generate candidate itemsets from frequentPatterns

        :param frequentPatterns: set of all frequent patterns to generate candidate patterns of each of size is length
        :type frequentPatterns: list
        :param length: size of each candidate patterns to be generated
        :type length: int
        :return: list of candidate patterns
        :rtype: list
        """
        candidates = list(_ab._c(frequentPatterns, 2))
        candidates = set([tuple(set(item[0]).union(set(item[1]))) for item in [x for x in candidates]])
        candidates = list({item for item in candidates if len(item) == length})
        return candidates

    def _genFrequentItems(self, database):
        """
        Get frequent items which length is 1

        :param database: database
        :type database: RDD
        :return: frequent items which length is 1
        :rtype: dict
        """
        frequentItems = dict(database.flatMap(lambda x: [(item, 1) for item in x])
                             .reduceByKey(lambda x, y: x + y)
                             .filter(lambda c: c[1] >= self._minSup)
                             .collect())
        return frequentItems

    def _getAllFrequentPatterns(self, database, frequentItems):
        """
        Get all frequent patterns and save them to self.oFile

        :param database: database
        :type : RDD
        :param frequentItems: dict
        :return:
        """

        # Get candidate patterns that length is 2
        candidates = list(_ab._c(frequentItems.keys(), 2))
        length = 3
        while len(candidates) != 0:
            # if each itemset of candidates is in each transaction, then create (itemset,1)
            mappedDatabase = database.flatMap(lambda transaction: self._Mapper(transaction, candidates))

            # aggregate the values by key by reduceByKey() method
            frequentPatterns = dict(
                mappedDatabase.reduceByKey(lambda x, y: x + y).filter(lambda c: c[1] >= self._minSup).collect())
            self._finalPatterns.update(frequentPatterns)
            candidates = self._genCandidateItemsets(list(frequentPatterns.keys()), length)
            length += 1

    def startMine(self):
        """
        Frequent pattern mining process will start from here

        :return:
        """
        self._startTime = _ab._time.time()

        # setting SparkConf and SparkContext to process in parallel
        conf = _ab._SparkConf().setAppName("parallelApriori").setMaster("local[*]")
        sc = _ab._SparkContext(conf=conf)

        # read database from iFile
        database = sc.textFile(self._iFile, self._numWorkers).map(lambda x: {int(y) for y in x.rstrip().split(self._sep)})

        # Calculating minSup as a percentage
        # self._minSup = database.count() * self._minSup / 100
        oneFrequentItems = self._genFrequentItems(database)
        self._finalPatterns = oneFrequentItems
        self._getAllFrequentPatterns(database, oneFrequentItems)

        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss

        sc.stop()


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 5 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 6:
            _ap = parallelApriori(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:
            _ap = parallelApriori(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        _ap.startMine()
        _finalPatterns = _ap.getPatterns()
        print("Total number of Frequent Patterns:", len(_finalPatterns))
        _ap.savePatterns(_ab._sys.argv[2])
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")
